@@ -1,12 +1,19 @@
 /*!
 # fltk-webview
 
-This provides webview functionality for embedded fltk windows. This currently works on Windows:
+This provides webview functionality for embedded fltk windows:
 
 ## Usage
+Add fltk-webview to your fltk application's Cargo.toml file:
+```toml
+[dependencies]
+fltk = "1"
+fltk-webview = "0.1"
+```
 
-```rust,no_run
-use fltk::{app, enums::Event, prelude::*, window};
+Then you can embed a webview using fltk_webview::Webview::create:
+```rust
+use fltk::{app, prelude::*, window};
 
 fn main() {
     let _app = app::App::default();
@@ -17,22 +24,28 @@ fn main() {
         .with_size(790, 590)
         .center_of_parent();
     win.end();
+    win.make_resizable(true);
     win.show();
 
-    // close the app when the main window is closed
-    win.set_callback(|_| {
-        if app::event() == Event::Close {
-            std::process::exit(0);
-        }
-    });
-
     let mut wv = fltk_webview::Webview::create(false, &mut wv_win);
     wv.navigate("https://google.com");
-
+    
     // the webview handles the main loop
     wv.run();
 }
 ```
+
+## Dependencies
+- fltk-rs's dependencies, which can be found [here](https://github.com/fltk-rs/fltk-rs#dependencies).
+- On Linux, webkit2gtk:
+    - Debian-based distros: `sudo apt-get install libwebkit2gtk-4.0-dev`.
+    - RHEL-based distros: `sudo dnf install webkit2gtk3-devel`.
+
+## Limitations
+- On windows, webview requires winrt headers, that means it's basically buildable with the MSVC toolchain. For Msys2/mingw, there are efforts to provide such headers, but nothing yet upstream.
+- On macos, need help with getting mouse input to work.
+- On linux, need help with Gnome's mutter window manager fighting for ownership of the webview window!
+
 */
 
 // Uses code from https://github.com/webview/webview_rust/blob/dev/src/webview.rs
@@ -46,9 +59,7 @@ use std::{
 };
 use webview_official_sys as wv;
 
-#[cfg(target_os = "macos")]
-#[macro_use]
-extern crate objc;
+static mut RUNNING: bool = true;
 
 pub(crate) trait FlString {
     fn safe_new(s: &str) -> CString;
@@ -109,23 +120,35 @@ impl Webview {
                     debug as i32,
                     &mut win.raw_handle() as *mut *mut raw::c_void as *mut raw::c_void,
                 );
+                win.draw(move |w| wv::webview_set_size(inner, w.w(), w.h(), 0));
+                win.parent().unwrap().set_callback(|_| {
+                    if app::event() == enums::Event::Close {
+                        RUNNING = false;
+                        std::process::exit(0);
+                    }
+                });
             }
             #[cfg(target_os = "macos")]
             {
-                use objc::runtime::*;
-                let win_view: *mut Object = msg_send![win.raw_handle() as *mut Object, contentView];
-                inner = wv::webview_create(debug as i32, std::ptr::null_mut());
-                let inner_win = wv::webview_get_window(inner);
-                wv::webview_set_size(inner, win.w(), win.h(), 0);
-                let inner_view: *mut Object = msg_send![inner_win as *mut Object, contentView];
-                let _: () = msg_send![inner_view, removeFromSuperview];
-                let _: () = msg_send![win_view, addSubview:inner_view positioned:1 relativeTo:0];
-                let _: () = msg_send![inner_view, acceptsFirstResponder];
-                let _: () = msg_send![inner_win as *mut Object, close];
+                pub enum NSWindow {}
+                extern "C" {
+                    pub fn make_delegate(child: *mut NSWindow, parent: *mut NSWindow);
+                }
+                let handle = win.raw_handle();
+                inner = wv::webview_create(debug as i32, handle as _);
+                make_delegate(wv::webview_get_window(inner) as _, handle as _);
+                win.draw(move |w| wv::webview_set_size(inner, w.w(), w.h(), 0));
+                win.parent().unwrap().set_callback(|_| {
+                    if app::event() == enums::Event::Close {
+                        RUNNING = false;
+                        std::process::exit(0);
+                    }
+                });
             }
             #[cfg(target_os = "linux")]
             {
                 use std::os::raw::*;
+                use signal_hook::{consts::signal::SIGINT, iterator::Signals};
                 pub enum GdkWindow {}
                 pub enum GtkWindow {}
                 pub enum Display {}
@@ -133,16 +156,17 @@ impl Webview {
                     pub fn gtk_init(argc: *mut i32, argv: *mut *mut c_char);
                     pub fn my_get_win(wid: *mut GtkWindow) -> *mut GdkWindow;
                     pub fn my_get_xid(w: *mut GdkWindow) -> u64;
-                    pub fn XReparentWindow(
-                        display: *mut Display,
-                        w: u64,
-                        parent: u64,
-                        x: i32,
-                        y: i32,
-                    );
-                    pub fn XMapWindow(display: *mut Display, w: u64);
-                    pub fn XFlush(disp: *mut Display);
+                    pub fn x_init(disp: *mut Display, child: u64, parent: u64);
+                    pub fn x_reparent(disp: *mut Display, child: u64, parent: u64);
+                    pub fn wv_at_exit(disp: *mut Display);
                 }
+                let mut signals = Signals::new(&[SIGINT]).unwrap();
+                std::thread::spawn(move || {
+                    for _sig in signals.forever() {
+                        wv_at_exit(app::display() as _);
+                        RUNNING = false;
+                    }
+                });
                 gtk_init(&mut 0, std::ptr::null_mut());
                 inner = wv::webview_create(debug as i32, std::ptr::null_mut() as _);
                 assert!(!inner.is_null());
@@ -151,12 +175,21 @@ impl Webview {
                 assert!(!temp.is_null());
                 let xid = my_get_xid(temp as _);
                 let flxid = win.raw_handle();
-                win.draw(move |w| {
-                    XMapWindow(app::display() as _, xid);
-                    XReparentWindow(app::display() as _, xid, flxid, w.x(), w.y());
-                    XFlush(app::display() as _);
-                    wv::webview_set_size(inner, w.w(), w.h(), 0);
+                if has_program("gnome-shell") {
+                    app::add_idle(move || {
+                        x_reparent(app::display() as _, xid, flxid);
+                        app::sleep(0.03);
+                    });
+                } else {
+                    x_init(app::display() as _, xid, flxid);
+                }
+                win.parent().unwrap().set_callback(|_| {
+                    if app::event() == enums::Event::Close {
+                        wv_at_exit(app::display() as _);
+                        RUNNING = false;
+                    }
                 });
+                win.draw(move |w| wv::webview_set_size(inner, w.w(), w.h(), 0));
             }
         }
         assert!(!inner.is_null());
@@ -254,12 +287,11 @@ impl Webview {
             #[cfg(not(any(target_os = "macos", target_os = "windows")))]
             {
                 extern "C" {
-                    pub fn gtk_main_iteration() -> i32;
+                    pub fn gtk_main_iteration_do(val: bool) -> bool;
+                }
+                while gtk_main_iteration_do(true) && RUNNING {
+                    app::check();
                 }
-                app::add_idle(|| {
-                    gtk_main_iteration();
-                });
-                app::run().unwrap();
             }
             #[cfg(any(target_os = "macos", target_os = "windows"))]
             wv::webview_run(*self.inner) 
@@ -271,3 +303,11 @@ impl Webview {
         unsafe { wv::webview_set_size(*self.inner, width, height, hints as i32) }
     }
 }
+
+#[cfg(target_os = "linux")]
+fn has_program(prog: &str) -> bool {
+    match std::process::Command::new(prog).arg("--version").output() {
+        Ok(out) => !out.stdout.is_empty(),
+        _ => false,
+    }
+}
